/*
 * "Intruder"-Robot-Tracking
 * A. Schiffler, 04-07/99 
 *
 * Requires XWindows running at 16bit depth 565 weight.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "bgrab.h"
#include "xutil.h"

/* Which kind of audio to use */

#ifdef USE_AUDIO
 #include "sounder.h"
#endif
 
#include "netsocket.h"

/* Grabbing and screen geometry */

#define WIDTH	640
#define HEIGHT	480
#define DEPTH 	2

/* Number of test images to be saved in test mode */

#define NUM_CAPTUREBUFFER	16

/* Where to store the defaults */

#define DEFAULT_FILENAME	"defaults.txt"

/* How many times to repeat reset sequence */

#define RESET_REPEAT_COUNT 	6

/* How often to send a reset */

#define RESET_EXECUTE_COUNT	(25*60*10)
int reset_counter=0;

/* Maximum length of text lines */

#define LINE_LEN	120
char blankline[81]="                                                                              ";

/* Settings (defaults) */

/* Display toggle */

int display_on=1;

/* Active area for processing */
int areamode=1;
int xmin=0;
int xmax=WIDTH-1;
int ymin=0;
int ymax=HEIGHT-1;
int xmid;
int ymid;

/* Current framegrabber settings */
int contrast;
int brightness;

/* Brightness level filter */
float limit = 0.5;

/* Audio play-object */
#ifdef USE_AUDIO
 struct sounder sound;
#endif

/* Audio control */
int mute_audio=0;

/* Frequencies (in Hz) */
#define BEEP_RESET	5000

int BEEP_CLOCK[3] = {7000,9000,11000};

/* Timing (in ms) */

// Current set
#define NUM_SET		2
#define TRACK_FRAME	1
#define RESET_START	5
#define RESET_DURATION	9
#define SET_START	15
#define SET_DURATION	10

/*
// 1sec scan, working
#define NUM_SET		2
#define TRACK_FRAME	1
#define RESET_START	0
#define RESET_DURATION	10
#define SET_START	10
#define SET_DURATION	10
*/

/*
// 1.5sec scan
#define NUM_SET		3
#define TRACK_FRAME	2
#define RESET_START	0
#define RESET_DURATION	10
#define SET_START	25
#define SET_DURATION	10
*/

/*
// 2sec scan
#define NUM_SET		4
#define TRACK_FRAME	2
#define RESET_START	10
#define RESET_DURATION	10
#define SET_START	30
#define SET_DURATION	10
*/

/* Quality control */

#define dlimit	50.0
#define dokmin	5.0
#define dokmax	10.0

/* Network settings */

int socket_port=1441;

/* Tracking database */
#define MAX_TARGETS	12
#define MAX_GROUPS	3
#define GROUP_TARGETS	4

#define AUDIO_BUFFER_BYTES	((2*44100/25)*(MAX_TARGETS*NUM_SET))

/* Thread-shared data block - lock/unlock during use */

struct target_info {
			/* Position and direction of scan */
			float posx;
			float posy;
			float direction;
			
			/* Quality of scan */
			float quality;
			
			/* Time of calculation */
			double time_updated;
			
			/* Has been updated since last read */
			int is_updated;
			
		} targets[MAX_TARGETS];
int current_frame=0;

/* Drawing tables and routines */

static float COS_TABLE[360], SIN_TABLE[360];

void InitializeSinCosTable(void)
{
 int i;
 double angle;
 
 for(i=0;i<360;i++) {
  angle=(double)i/180.0*M_PI;
  COS_TABLE[i]=cos(angle);
  SIN_TABLE[i]=sin(angle);
 }
}

#ifdef USE_AUDIO
void CalculateWaveform(struct sounder *sound)
{
 int i,j,k,l;
 int frequency;
 int frame;
 int frameoffset;
 int samples;
 unsigned char *curbuffer;
#ifdef WAVELOG
 unsigned short *curbuffer2;
#endif
 double angle;
 double anglestep;
 double envangle;
 double envanglestep;
 double dvalue;
 double evalue;
 int ivalue;
 
#ifdef DEBUG
 fprintf (stderr,"Calculating waveform\n");
#endif

 /* Clear complete buffers */
 memset (sound->audio_buffer,0,sound->audio_buffer_size);
 memset (sound->audio_buffer_reset,0,sound->audio_buffer_size);

#ifdef TESTAUDIO
  /* Fill buffer with single frequency */
  frequency=TESTFREQ;
  curbuffer=(unsigned char *)&sound->audio_buffer[0];
  angle=0.0;
  anglestep=2.0*M_PI*((double)frequency/44100.0);
  for (j=0; j<(sound->audio_buffer_size/2); j++) {
   dvalue=32767.0*(0.5*(1.0+sin(angle)));
   ivalue=dvalue;
   *curbuffer=(unsigned char)(ivalue & 0xff);
   curbuffer++;
   *curbuffer=(unsigned char)((ivalue >> 8) & 0xff);
   curbuffer++;
   angle += anglestep;
  }
#else 

 /* Normal buffer */
 
 /* Write reset pulses into buffer */
 for (i=0; i<MAX_GROUPS; i++) {
  /* Set frequency and work out angle increments */
  frequency=BEEP_RESET;
  anglestep=2.0*M_PI*((double)frequency/44100.0);
  /* Which frame is this pulse in */
  frame=NUM_SET*i*GROUP_TARGETS;
  frameoffset=frame*(2*1764);
  /* Adjust offset for start of pulse */
  frameoffset += (2*(int)((float)RESET_START*44.1));
  /* Work out pulse length in samples */
  samples = (int)((float)RESET_DURATION*44.1);
  curbuffer=(unsigned char *)&sound->audio_buffer[frameoffset];
  /* Work out envelope settings */
  envanglestep=M_PI/(double)samples;
  /* Fill this buffer section */
  angle=0.0;
  envangle=0.0;
  for (j=0; j<samples; j++) {
   evalue=sin(envangle);
   for (l=0; l<i; l++) {
    evalue=sqrt(evalue);
   } 
   dvalue=32767.0*(0.5*(1.0+sin(angle)))*evalue;
   ivalue=dvalue;
   *curbuffer=(unsigned char)(ivalue & 0xff);
   curbuffer++;
   *curbuffer=(unsigned char)((ivalue >> 8) & 0xff);
   curbuffer++;
   angle += anglestep;
   envangle += envanglestep;
  }
 }
  
 /* Write set pulses into buffer */
 for (i=0; i<MAX_GROUPS; i++) {
  for (k=0; k<GROUP_TARGETS; k++) { 
   /* Set frequency and work out angle increments */
   frequency=BEEP_CLOCK[i];
   anglestep=2.0*M_PI*((double)frequency/44100.0);
   /* Which frame is this pulse in */
   frame=NUM_SET*(i*GROUP_TARGETS+k);
   frameoffset=frame*(2*1764);
   /* Adjust offset for start of pulse */
   frameoffset += (2*(int)((float)SET_START*44.1));
   /* Work out pulse length in samples */
   samples = (int)((float)SET_DURATION*44.1);
   curbuffer=(unsigned char *)&sound->audio_buffer[frameoffset];
   /* Work out envelope settings */
   envanglestep=M_PI/(double)samples;
   /* Fill this buffer section */
   angle=0.0;
   envangle=0.0;
   for (j=0; j<samples; j++) {
    evalue=sin(envangle);
    for (l=0; l<i; l++) {
     evalue=sqrt(evalue);
    } 
    dvalue=32767.0*(0.5*(1.0+sin(angle)))*evalue;
    ivalue=dvalue;
    *curbuffer=(unsigned char)(ivalue & 0xff);
    curbuffer++;
    *curbuffer=(unsigned char)((ivalue >> 8) & 0xff);
    curbuffer++;
    angle += anglestep;
    envangle += envanglestep;
   }
  } 
 }

 /* Reset buffer - no reset pulses in this one, repeat a couple of times */
 
 /* Write set pulses into buffer */
 for (i=0; i<MAX_GROUPS; i++) {
  for (k=0; k<GROUP_TARGETS; k++) { 
   /* Set frequency and work out angle increments */
   frequency=BEEP_CLOCK[i];
   anglestep=2.0*M_PI*((double)frequency/44100.0);
   /* Which frame is this pulse in */
   frame=NUM_SET*(i*GROUP_TARGETS+k);
   frameoffset=frame*(2*1764);
   /* Adjust offset for start of pulse */
   frameoffset += (2*(int)((float)SET_START*44.1));
   /* Work out pulse length in samples */
   samples = (int)((float)SET_DURATION*44.1);
   curbuffer=(unsigned char *)&sound->audio_buffer_reset[frameoffset];
   /* Work out envelope settings */
   envanglestep=M_PI/(double)samples;
   /* Fill this buffer section */
   angle=0.0;
   envangle=0.0;
   for (j=0; j<samples; j++) {
    evalue=sin(envangle);
    for (l=0; l<i; l++) {
     evalue=sqrt(evalue);
    } 
    dvalue=32767.0*(0.5*(1.0+sin(angle)))*evalue;
    ivalue=dvalue;
    *curbuffer=(unsigned char)(ivalue & 0xff);
    curbuffer++;
    *curbuffer=(unsigned char)((ivalue >> 8) & 0xff);
    curbuffer++;
    angle += anglestep;
    envangle += envanglestep;
   }
  } 
 }
#endif   // TESTAUDIO

#ifdef WAVELOG
 curbuffer2=(unsigned short *)sound->audio_buffer;
 for (i=0; i<(sound->audio_buffer_size/2); i++) {
  printf ("%i %i\n",i,*curbuffer2);
  curbuffer2++;
 }
#endif  // WAVELOG
}
#endif   // USE_AUDIO

void hline (unsigned short *image, int x1, int x2, int y)
{
 unsigned short *curpos;
 int dummy;
 
 if (x2<x1) {
  dummy=x1;
  x1=x2;
  x2=dummy;
 }
 if (x1<0) {
  x1=0;
 } else if (x1>(WIDTH-1)) {
  x1=WIDTH-1;
 } 
 if (x2<0) {
  x2=0;
 } else if (x2>(WIDTH-1)) {
  x2=WIDTH-1;
 } 
 if (y<0) {
  y=0;
 } else if (y>(HEIGHT-1)) {
  y=HEIGHT-1;
 } 
 curpos = image;
 curpos += (x1+y*WIDTH);
 memset(curpos,0xff,2*(x2-x1+1));
}

void vline (unsigned short *image, int x, int y1, int y2)
{
 int i;
 unsigned short *curpos;
 int dummy;
 
 if (y2<y1) {
  dummy=y1;
  y1=y2;
  y2=dummy;
 }
 if (x<0) {
  x=0;
 } else if (x>(WIDTH-1)) {
  x=WIDTH-1;
 } 
 if (y1<0) {
  y1=0;
 } else if (y1>(HEIGHT-1)) {
  y1=HEIGHT-1;
 } 
 if (y2<0) {
  y2=0;
 } else if (y2>(HEIGHT-1)) {
  y2=HEIGHT-1;
 } 
 curpos = image;
 curpos += (x+y1*WIDTH);
 for (i=y1; i<=y2; i++) {
  *curpos=0xffff;
  curpos += WIDTH;
 }
}

/* Color tracking table */

unsigned char conversion_table[64*1024];

void InitializeConversionTable(void)
{
 int i;
 unsigned short us;
 float rf,gf,bf,yf;
  
 for (i=0; i<(64*1024); i++) {
  us=(unsigned short)i;
  rf=(float)(us >> 11);
  rf *= (1.0/31.0);
  gf=(float)((us >> 5) & 63);
  gf *= (1.0/63.0);
  bf=(float)(us & 31);
  bf *= (1.0/31.0);
  yf =  (0.257 * rf) + (0.504 * gf) + (0.098 * bf);
  conversion_table[i]=(yf>limit);
 }
}

void SaveDefaultSettings()
{
 FILE *deffile;
 
 deffile=fopen(DEFAULT_FILENAME,"w");

 fprintf(deffile,"# Framegrabber controls\n");
 fprintf(deffile,"contrast	%i\n",contrast);
 fprintf(deffile,"brightness	%i\n",brightness);
 fprintf(deffile,"# Active area definition\n");
 fprintf(deffile,"xmin	%i\n",xmin);
 fprintf(deffile,"xmax	%i\n",xmax);
 fprintf(deffile,"ymin	%i\n",ymin);
 fprintf(deffile,"ymax	%i\n",ymax);
 fprintf(deffile,"# Audio on/off\n");
 fprintf(deffile,"mute	%i\n",mute_audio);
 fprintf(deffile,"# Brightness limit (in percent)\n");
 fprintf(deffile,"limit	%i\n",(int)(100.0*limit));
 fprintf(deffile,"# Display On/Off\n");
 fprintf(deffile,"display %i\n",display_on);
 
 fclose(deffile);
}

void LoadDefaultSettings()
{
 FILE *deffile;
 char aline[LINE_LEN];
 char tag[LINE_LEN];
 int value;
 
 /* Open file */
 if ((deffile=fopen(DEFAULT_FILENAME,"r"))!=NULL) {
 
  /* Scan file for tags/value pairs */
  /* Fetch a line */
  fgets(aline,LINE_LEN,deffile);  
  while (!feof(deffile)) {
   /* Ignore line starting with a # */
   if (aline[0]!='#') {
    sscanf(aline,"%s %i",tag,&value);
#ifdef DEBUG   
    fprintf (stderr,"Tag=%s Value=%i\n",tag,value);
#endif    
    if (strncasecmp(aline,"contrast",8)==0) {
     contrast=value;
    }
    if (strncasecmp(aline,"brightness",8)==0) {
     brightness=value;
    }
    if (strncasecmp(aline,"xmin",4)==0) {
     xmin=value;
    }
    if (strncasecmp(aline,"xmax",4)==0) {
     xmax=value;
    }
    if (strncasecmp(aline,"ymin",4)==0) {
     ymin=value;
    }
    if (strncasecmp(aline,"ymax",4)==0) {
     ymax=value;
    }
    if (strncasecmp(aline,"mute",4)==0) {
     mute_audio=value;
    }
    if (strncasecmp(aline,"limit",5)==0) {
     limit=0.01*(float)value;
    }
    if (strncasecmp(aline,"display",7)==0) {
     display_on=(value & 1);
    }
   }  
   /* Get next line */
   fgets(aline,LINE_LEN,deffile);  
  }
 
  /* Done - now close */
  fclose(deffile);

 } 
}

void main (int argc, char *argv[])
{
 int x,y,i,j;
 struct fgdevice fg;
 struct xwinbuffer xwin;
 struct netthread nt;
 int not_first_time=0;
 int in_loop=1;
 KeySym key;
 int button;
 int mousex,mousey;
 int current_target=0;
 unsigned short *curpos;
 long sumn;
 long sumx;
 long sumy;
 int x1,x2,y1,y2;
 float nfactor,screenx,screeny,relx,rely,dir;
 int screenxi,screenyi;
 float dx,dy,d,dmax;
 float quality;
 int foundx,foundy;
 int captureflag=0;
 unsigned short *capturebuffer[NUM_CAPTUREBUFFER];
 char cfilename[LINE_LEN];
 FILE *capturefile;
 int rval,gval,bval;
 unsigned char rgbtriple[3];
 float curfps=0.0;
#ifdef USE_AUDIO
 struct sounder sound;
#endif

 /* Clear text screen */
 printf ("\033[H");
 for (i=0; i<30; i++) {
  printf ("%s\n",blankline);
 }
   
 /* Setup framegrabber */  
 if (argc>1) {
  if (fg_open_device (&fg, argv[1])!=0) exit(-1); 	
 } else {
  if (fg_open_device (&fg, "/dev/video")!=0) exit(-1); 	
 } 
 if (fg_print_info (&fg)!=0) exit(-1);
 if (fg_set_channel (&fg,1)!=0) exit(-1);
 sleep(1);

 /* Clear text screen */
 printf ("\033[H");
 for (i=0; i<30; i++) {
  printf ("%s\n",blankline);
 }

 /* Read settings from card */
 contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
 brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;

 /* Load defaults */
 LoadDefaultSettings();

 /* Precalculate screen settings */
 xmid=xmin+(xmax-xmin+1)/2;
 ymid=ymin+(ymax-ymin+1)/2;

 /* Set settings */
 fg_set_setting(&fg,SETTING_CONTRAST,contrast);
 fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
 
 /* Setup image-binarization table for the first time */
 InitializeConversionTable();     

 /* Setup math table */
 InitializeSinCosTable();     
 
 /* Start grabbing */
 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_RGB565)!=0) exit(-1);
 fg_set_fps_interval(&fg,100);

 /* Create capturebuffers */
 for (i=0; i<NUM_CAPTUREBUFFER; i++) {
  capturebuffer[i]=(unsigned short *)malloc(fg.image_size);
 }
  
#ifdef USE_AUDIO 
 /* Setup soundcard */
 if (sound_init(&sound,"/dev/dsp",AUDIO_BUFFER_BYTES)!=0) exit(-1);
 /* Set mixer to maximum volume */
 if (mixer_init("/dev/mixer",100)!=0) exit(-1);
 /* Generate waveform for beep sequence */
 CalculateWaveform(&sound);
#endif
 
 /* Setup network */
 if (start_network_socket(&nt,socket_port,BE_VERBOSE)!=0) exit(-1);
 
 /* Setup screen */
 if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, DEPTH)!=0) exit(-1);
  
 /* Loop */
 while (in_loop) { 

  /* Update framebuffer with new image */
  if (fg_get_next_image(&fg)!=NULL) {
     
   /* Process test mode - capturing frame images */
   if (captureflag>0) {
    /* Wait for frame zero and update captureflag */
    if (current_frame==0) {
     captureflag++;
    }
    /* Start capturing if captureflag is in the right mode */     
    if (captureflag>1) {
     if (current_frame<NUM_CAPTUREBUFFER) {
      /* Copy image to capturebuffer */
      memcpy((char *)capturebuffer[current_frame],(char *)fg.current_image,fg.image_size);
     } 
     if (current_frame==(NUM_CAPTUREBUFFER-1)) {
      /* Save images */
      for (i=0; i<NUM_CAPTUREBUFFER; i++) {
       sprintf(cfilename,"capture%i.ppm",i);
       printf ("Writing %s ...\n",cfilename);
       capturefile=fopen(cfilename,"w");
       /* Write headers */
       fprintf(capturefile,"P6\n# intrudertrack\n%i %i\n255\n",WIDTH,HEIGHT);
       /* Dump screen */
       curpos=capturebuffer[i];
       for (y=0; y<HEIGHT; y++) {
        for (x=0; x<WIDTH; x++) {
         rval=(*curpos & 0xf800); rval = rval >> 8;
         gval=(*curpos & 0x07d0); gval = gval >> 3;
         bval=(*curpos & 0x001f); bval = bval << 3;
         rgbtriple[0]=(unsigned char)rval;
         rgbtriple[1]=(unsigned char)gval;
         rgbtriple[2]=(unsigned char)bval;
         fwrite ((void *)rgbtriple,3,1,capturefile);
         curpos++;
        }
       }
       fclose(capturefile);
      } 
      /* No more capturing */
      captureflag=0;
     } 
    } 
   }

   /* Check frame sequence to send triggers */
   switch (current_frame) {
    case 0:
#ifdef USE_AUDIO
     /* Play audio-seqence through soundcard */
     if (!mute_audio) {
      if (sound_me(&sound)!=0) exit(-1);
     } 
#endif
     break;

    case (NUM_SET*MAX_TARGETS): 
     /* This the syncronize frame - send nothing - dont process */ 
     break;

    default: 
               
      /* Process current targets image */
      /* First determine central moment and min/max values */
      sumn=0;
      sumx=0;
      sumy=0;
      x1=WIDTH;
      x2=0;
      y1=HEIGHT;
      y2=0;
      curpos=(unsigned short *)fg.current_image;
      for (y=0; y<HEIGHT; y++) {
       for (x=0; x<WIDTH; x++) {
        if ((x>=xmin) && (x<=xmax) && (y>=ymin) && (y<=ymax)) {
         if (conversion_table[*curpos]) {
          /* Add */
          sumn++;
          sumx += x;
          sumy += y;
          /* Store limits where we found pixels*/
          if (x<x1) x1=x;
          if (x>x2) x2=x;
          if (y<y1) y1=y;
          if (y>y2) y2=y;
          }
         } 
        curpos++;
       } 
      }

     /* This is a normal frame - process "track-frames" */
     if ((current_frame % NUM_SET) == TRACK_FRAME) {

      /* Calculate current target number */
      current_target=(current_frame/NUM_SET)-1;
      if (current_target<0) current_target=0;
      
      /* Check if we found any pixels at all */
      if (sumn>2) {
       /* Calculate COM in screen coordinates */
       nfactor=1.0/(float)sumn;
       screenx=(float)sumx*nfactor;
       screenxi=screenx;
       screeny=(float)sumy*nfactor;
       screenyi=screeny;
   
       /* Setup second scan within limits */
       curpos=(unsigned short *)fg.current_image;
       curpos += (x1+WIDTH*y1);
       dmax=0.0;
       foundx=screenxi;
       foundy=screenyi;
       
       /* Second: determine pixel that is farthest away from COM */
       if (display_on) {
        /* Calculation with screen update */
        for (y=y1; y<=y2; y++) {
         for (x=x1; x<=x2; x++) {
          if (conversion_table[*curpos]) {
           dx=(float)x-screenx;
           dy=(float)y-screeny;
           d=dx*dx+dy*dy;
           /* Sanity test */
           if (d<(dlimit*dlimit)) {
            /* Maximum test, store coordinate of farthest pixel */
            if (d>dmax) {
             foundx=x;
             foundy=y;
             dmax=d;
            }
           } 
           /* Mark on screen */
           *curpos=0xf800;
          }
          curpos++;
         }
         curpos += (WIDTH-x2+x1-1);
        }
       } else {
        /* Calculation without screen update */
        for (y=y1; y<=y2; y++) {
         for (x=x1; x<=x2; x++) {
          if (conversion_table[*curpos]) {
           dx=(float)x-screenx;
           dy=(float)y-screeny;
           d=dx*dx+dy*dy;
           /* Sanity test */
           if (d<(dlimit*dlimit)) {
            /* Maximum test, store coordinate of farthest pixel */
            if (d>dmax) {
             foundx=x;
             foundy=y;
             dmax=d;
            }
           } 
          }
          curpos++;
         }
         curpos += (WIDTH-x2+x1-1);
        }
       }

       /* Normalize COM position and calculate direction */
       /* X is normalized to the y-size */
       nfactor=1.0/(float)(ymax-ymin);
       relx=(screenx-(float)xmin)*nfactor;
       nfactor=1.0/(float)(ymax-ymin);
       rely=(screeny-(float)ymin)*nfactor;
       dir=(180.0/M_PI)*(float)atan2((double)((float)foundy-screeny),(double)((float)foundx)-screenx);
       if (dir<0.0) dir += 360.0;
   
       /* Determine quality value */
       if (dmax<(dokmin*dokmin)) {
        quality=(dmax/(dokmin*dokmin));
       } else if (dmax>(dokmax*dokmax)) {
        quality=(1.0-(dmax-(dokmax*dokmax))/((dlimit*dlimit)-(dokmax*dokmax)));
       } else if (dmax>(dlimit*dlimit)){
        quality=0.0;
       } else { 
        quality=1.0;
       } 

      } else {
        /* No pixels ... */
        relx=0.0;
        rely=0.0;
        dir=0.0;
       
        /* ... no quality */
        quality=0.0;
      }
      
      /* Store results */
      shared_data_lock(&nt);
      targets[current_target].posx=relx;
      targets[current_target].posy=rely;
      targets[current_target].direction=(360.0-dir);
      targets[current_target].quality=quality;
      targets[current_target].time_updated=timeGet();
      targets[current_target].is_updated=1;
      shared_data_unlock(&nt);
    } /* current_frame test */
   } /* current_frame switch */
   
   /* Update framecounter */
   current_frame++;
   if (current_frame==(NUM_SET*MAX_TARGETS+NUM_SET)) current_frame=0;
   
   if (display_on) {
    /* Show tracking area */
    hline((unsigned short *)fg.current_image,xmin,xmin+5,ymin);
    hline((unsigned short *)fg.current_image,xmin,xmin+5,ymax);
    hline((unsigned short *)fg.current_image,xmax-5,xmax,ymin);
    hline((unsigned short *)fg.current_image,xmax-5,xmax,ymax);
    vline((unsigned short *)fg.current_image,xmin,ymin,ymin+5);
    vline((unsigned short *)fg.current_image,xmax,ymin,ymin+5);
    vline((unsigned short *)fg.current_image,xmin,ymax-5,ymax);
    vline((unsigned short *)fg.current_image,xmax,ymax-5,ymax);
   
    /* Show tracking mode */
    if (areamode) {
     hline((unsigned short *)fg.current_image,xmid-5,xmid+5,ymid);
     vline((unsigned short *)fg.current_image,xmid,ymid-5,ymid+5);
    } else {
     hline((unsigned short *)fg.current_image,xmid-5,xmid+5,ymin);
     hline((unsigned short *)fg.current_image,xmid-5,xmid+5,ymax);
     vline((unsigned short *)fg.current_image,xmin,ymid-5,ymid+5);
     vline((unsigned short *)fg.current_image,xmax,ymid-5,ymid+5);
    }
   
    /* Show previous tracking positions */
    for (i=0; i<MAX_TARGETS; i++) {
      /* Get data */
      shared_data_lock(&nt);
      screenxi=xmin+targets[i].posx*(xmax-xmin);
      screenyi=ymin+targets[i].posy*(ymax-ymin);
      dir=targets[i].direction;
      shared_data_unlock(&nt);
      /* Adjust direction */
      dir = 360.0-dir;
      if (dir<0.0) dir += 360.0;
      if (dir>360.0) dir -= 360.0;
      /* Show position */
      x=screenxi;
      y=screenyi;
      hline((unsigned short *)fg.current_image,x-4,x+4,y);
      vline((unsigned short *)fg.current_image,x,y-4,y+4);
      /* Show direction */
      for (j=0; j<360; j += 36) {
       x=screenxi+(int)(12.0*COS_TABLE[j % 360]);
       y=screenyi+(int)(12.0*SIN_TABLE[j % 360]);
       if ((x>=0) && (x<WIDTH) && (y>=0) && (y<HEIGHT)) {
        curpos=(unsigned short *)fg.current_image;
        curpos += (x+y*WIDTH);
        *curpos=0xffff;
       } 
      } 
      x=screenxi+(int)(12.0*COS_TABLE[(int)dir % 360]);
      y=screenyi+(int)(12.0*SIN_TABLE[(int)dir % 360]);
      hline((unsigned short *)fg.current_image,x-1,x+1,y-1);
      hline((unsigned short *)fg.current_image,x-1,x+1,y);
      hline((unsigned short *)fg.current_image,x-1,x+1,y+1);
    }
   
    /* Copy image to screenbuffer */
    memcpy((char *)xwin.shmimage->data,(char *)fg.current_image,fg.image_size);

    /* Draw it */
    sync_xwinbuffer(&xwin);
   } 
  }
 
  /* Poll keyboard and mouse */
  if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {

    /* Print position on mouse-press */
    if (button) {
     printf ("\nX(screen)=%i Y(screen)=%i\n",mousex,mousey);
     if ((mousex>=xmin) && (mousex<=xmax) && (mousey>=ymin) && (mousey<=ymax)) {
      printf ("X(track)=%f Y(track)=%f\n",(float)(mousex-xmin)/(float)(xmax-xmin),(float)(mousey-ymin)/(float)(ymax-ymin));
     }    
     button=0;
    }
 
   switch (key) {
    
    case XK_q:     /* q - quit */
     in_loop=0;
     break; 
     
    case XK_s:     /* s - save settings */ 
     SaveDefaultSettings();
     fprintf (stderr,"Default settings saved.\n");
     break; 
    
    case XK_t:     /* t - capture frames and save them for tests */
     if (captureflag==0) {
      captureflag++;     
     }
     break;

    case XK_r:     /* r - trigger reset pulses */
     sound.do_reset=RESET_REPEAT_COUNT;
     break;

    case XK_d:     /* d - turn screen output on and off */
     display_on = (display_on+1) & 1;
     if (display_on==0) { 
      /* Clear screenbuffer */
      bzero((char *)xwin.shmimage->data,fg.image_size);
      /* Draw it */
      sync_xwinbuffer(&xwin);
     } 
     break;
     
    case XK_l:
     if (limit<1.0) {
      limit += 0.01;
      InitializeConversionTable();     
      fprintf (stderr,"Limit %3.0f %%\n",limit*100.0);
     } 
     break; 
    
    case XK_k:
     if (limit>0.0) {
      limit -= 0.01 ;
      InitializeConversionTable();     
      fprintf (stderr,"Limit %3.0f %%\n",limit*100.0);
     } 
     break; 
    case XK_c:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
     fprintf (stderr,"Contrast %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 
    case XK_x:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)+256;
     fprintf (stderr,"Contrast %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 
    case XK_b:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;
     fprintf (stderr,"Brightness %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 
    case XK_v:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)+256;
     fprintf (stderr,"Brightness %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 
    case XK_m:
     mute_audio=((mute_audio+1) & 1);
     if (mute_audio) {
      fprintf (stderr,"Audio OFF.\n");
     } else {
      fprintf (stderr,"Audio ON.\n");
     }
     break; 
    case XK_Home:
     areamode=((areamode + 1) % 2);
     if (areamode) {
      fprintf (stderr,"Cursors Enlarge.\n");
     } else {
      fprintf (stderr,"Cursors Shift.\n");
     }
     break; 
    case XK_Left:
     if (areamode) {
      if (xmin>0) {
       xmin--;
       xmax--;
      }
     } else {
      if ((xmax-xmin)>2) {
       xmin++;
       xmax--;
      } 
     }
     xmid=xmin+(xmax-xmin+1)/2;
     ymid=ymin+(ymax-ymin+1)/2;
     break; 
    case XK_Right:
     if (areamode) {
      if (xmax<(WIDTH-1)) {
       xmin++;
       xmax++;
      }
     } else {
      if ((xmin>0) && (xmax<(WIDTH-1))) {
       xmin--;
       xmax++;
      } 
     }
     xmid=xmin+(xmax-xmin+1)/2;
     ymid=ymin+(ymax-ymin+1)/2;
     break; 
    case XK_Up:
     if (areamode) {
      if (ymin>0) {
       ymin--;
       ymax--;
      }
     } else {
      if ((ymax-ymin)>2) {
       ymin++;
       ymax--;
      } 
     }
     xmid=xmin+(xmax-xmin+1)/2;
     ymid=ymin+(ymax-ymin+1)/2;
     break; 
    case XK_Down:
     if (areamode) {
      if (ymax<(HEIGHT-1)) {
       ymin++;
       ymax++;
      }
     } else {
      if ((ymin>0) && (ymax<(HEIGHT-1))) {
       ymin--;
       ymax++;
      } 
     }
     xmid=xmin+(xmax-xmin+1)/2;
     ymid=ymin+(ymax-ymin+1)/2;
     break; 
   }
  }
    
#ifndef DEBUG
  /* Text status display */

  /* Clear text screen and blank it */
  printf ("\033[H");

  /* Write title */
  printf ("INTRUDERTRACK - A. Schiffler, 04-07/99, ZKM\n");
  
  /* Output sequence # */   
  printf ("Current frame : %2i\n",current_frame);

  /* Determine reset state */
  reset_counter++;
  if (reset_counter>RESET_EXECUTE_COUNT) {
   sound.do_reset=RESET_REPEAT_COUNT;
   reset_counter=0;
  }
  
  /* Print fps */
  if (fg.framecount==0) {
   if (not_first_time) {
    curfps=fg_get_fps(&fg);
   } else {
    not_first_time=1; 
   } 
  } 
  printf ("FPS ......... : %.2f\n",curfps);
  printf ("Capture...... : %2i\n",captureflag);
  printf ("Xsize ....... : %i\n",xmax-xmin);
  printf ("Ysize ....... : %i\n",ymax-ymin);
  /* Print robot data */
  shared_data_lock(&nt);
  for (i=0; i<MAX_TARGETS; i++) {
    printf ("Robot %2i .... : %5.3f %5.3f %3.0f %.3f %3.0f%% %1i\n",
    					i,
  					targets[i].posx,
  					targets[i].posy,
  					targets[i].direction,
  					targets[i].time_updated,
  					targets[i].quality*100.0,
  					targets[i].is_updated
    );
  }
  shared_data_unlock(&nt);
#endif
  
 }           

 /* Shutdown graphcis */
 if (close_xwinbuffer(&xwin)!=0) exit(-1);

 /* Shutdown network */
 if (stop_network_socket(&nt)!=0) exit(-1);
 
#ifdef USE_AUDIO
 /* Shutdown sound */
 if (sound_close(&sound)!=0) exit(-1);
#endif 
 /* Shutdown framegrabber */
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	
 if (fg_close_device (&fg)!=0) exit(-1); 	
 
 exit(0);
}

int command_to_output(char *inbuffer, char* outbuffer, int outbufferlen)
{
 int target_selection;
 int target_reference;
 
#ifdef DEBUG
 fprintf (stderr,"Socket Read: %s\n",inbuffer);
#endif

 if ((inbuffer[0]=='t' ) || (inbuffer[0]=='T')) {
  /* System time */
  snprintf (outbuffer,outbufferlen,"%.2f\n",timeGet());
  return(0);
 } else if ((inbuffer[0]=='r' ) || (inbuffer[0]=='R')) {
  /* Trigger reset */
  sound.do_reset=RESET_REPEAT_COUNT;
  snprintf (outbuffer,outbufferlen,"%i\n",RESET_REPEAT_COUNT);
  return(0);
 } else {
  /* Read number */
  target_selection=0;
  sscanf (inbuffer,"%i",&target_selection);
  
  /* Convert input command to output string */
  if ((target_selection>0) && (target_selection<=MAX_TARGETS)) {
   target_reference=target_selection-1;
   snprintf (outbuffer,outbufferlen,"%2i %5.3f %5.3f %5.3f %5.3f %.2f %1i\n",
  					target_selection,
  					targets[target_reference].posx,
  					targets[target_reference].posy,
  					targets[target_reference].direction,
  					targets[target_reference].quality,
  					targets[target_reference].time_updated,
  					targets[target_reference].is_updated
  					);
#ifdef DEBUG
 fprintf (stderr,"Socket Write: %s",outbuffer);
#endif  					
   targets[target_reference].is_updated=0;					
   return(0);
  }
 }
 
 return(-1);
}
